#include "ChipArp.h"



AudioEffect* createEffectInstance(audioMasterCallback audioMaster)
{
	return new ChipArp(audioMaster);
}



ChipArp::ChipArp(audioMasterCallback audioMaster) : AudioEffectX(audioMaster, NUM_PROGRAMS, NUM_PARAMS)
{
	setNumInputs(NUM_INPUTS);
	setNumOutputs(NUM_OUTPUTS);
	setUniqueID(PLUGIN_UID);

	isSynth();
	canProcessReplacing();

	strcpy(programName,"Default"); // default program name 

	pBPM=1.0f/MAX_BPM*120.0f;
	pSpeed=.5f;
	pDirection=0;
	pBendRange=0;

	arpBaseNote=-1;			//no note is playing
	arpCurrentNote=-1;		//no arpeggio is playing
	arpCurrentDir=1;		//up
	arpCounterAcc=0;
	arpCounterAdd=0;

	bendOutRange=0;

	bendIn=0;
	bendInRange=2.0f;

	MidiRPNLSB=0;
	MidiRPNMSB=0;
	MidiDataLSB=0;
	MidiDataMSB=0;

	memset(MidiKeyState,0,sizeof(MidiKeyState));

	MidiKeyDownCount=0;

	MidiInQueue.clear();

	suspend();
}



ChipArp::~ChipArp()
{
}



void ChipArp::setParameter(VstInt32 index,float value)
{
	switch(index)
	{
	case pIdBPM:		pBPM      =value; break;
	case pIdSpeed:		pSpeed    =value; break;
	case pIdDirection:	pDirection=value; break;
	case pIdBendRange:	pBendRange=value; break;
	}
}



float ChipArp::getParameter(VstInt32 index)
{
	switch(index)
	{
	case pIdBPM:		return pBPM;
	case pIdSpeed:		return pSpeed;
	case pIdDirection:	return pDirection;
	case pIdBendRange:	return pBendRange;
	}

	return 0;
} 



void ChipArp::getParameterName(VstInt32 index,char *label)
{
	switch(index)
	{
	case pIdBPM:		strcpy(label,"BPM");		break;
	case pIdSpeed:		strcpy(label,"Speed");		break;
	case pIdDirection:	strcpy(label,"Direction");	break;
	case pIdBendRange:	strcpy(label,"Bend range"); break;
	default:			strcpy(label,"");
	}
} 



void ChipArp::getParameterDisplay(VstInt32 index,char *text)
{
	switch(index)
	{
	case pIdBPM:		float2string(pBPM*MAX_BPM,text,5); break;
	case pIdSpeed:		sprintf(text,"%i",pSpeedSteps[(int)(pSpeed*7.99f)]); break;
	case pIdDirection:	strcpy(text,pArpDirNames[(int)(pDirection*2.99f)]); break;
	case pIdBendRange:	strcpy(text,pBendRangeNames[(int)(pBendRange*4.99f)]); break;
	default:			strcpy(text,"");
	}
} 



void ChipArp::getParameterLabel(VstInt32 index,char *label)
{
	switch(index)
	{
	case pIdBPM:		strcpy(label,"BPM"); break;
	case pIdSpeed:		strcpy(label,"x"); break;
	default:			strcpy(label,"");
	}
} 



VstInt32 ChipArp::getProgram(void)
{
	return 0;
}



void ChipArp::setProgram(VstInt32 program)
{
}



void ChipArp::getProgramName(char* name)
{
	strcpy(name,programName); 
}



VstInt32 ChipArp::canDo(char* text)
{
	if(!strcmp(text,"receiveVstEvents"   )) return 1;
	if(!strcmp(text,"receiveVstMidiEvent")) return 1;
	if(!strcmp(text,"sendVstEvents")) return 1;
	if(!strcmp(text,"sendVstMidiEvent")) return 1;

	return -1;
}



VstInt32 ChipArp::getNumMidiInputChannels(void)
{
	return 1;
}



VstInt32 ChipArp::getNumMidiOutputChannels(void)
{
	return 1;
}



void ChipArp::setProgramName(char* name)
{
	strcpy(programName,name); 
} 



VstInt32 ChipArp::processEvents(VstEvents* ev) 
{ 
	VstMidiEvent event;
	VstInt32 i;

	for(i=0;i<ev->numEvents;++i) 
	{ 
		memcpy(&event,(VstMidiEvent*)ev->events[i],sizeof(VstMidiEvent));

		MidiInQueue.push_back(event);
	}

	return 1; 
} 



void ChipArp::SetPitchBendRange(void)
{
	VstEvents evs;
	VstMidiEvent evm;

	evs.numEvents=1;

	evm.byteSize=sizeof(evm);
	evm.deltaFrames=0;
	evm.type=kVstMidiType;

	evm.midiData[0]=(char)0xb0;	//RPN LSB
	evm.midiData[1]=0x64;
	evm.midiData[2]=0x00;

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);

	evm.midiData[0]=(char)0xb0;	//RPN MSB
	evm.midiData[1]=0x65;
	evm.midiData[2]=0x00;

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);

	evm.midiData[0]=(char)0xb0;	//DATA MSB, Pitch Bend Range
	evm.midiData[1]=0x06;
	evm.midiData[2]=(char)(bendOutRange*2);

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);

	evm.midiData[0]=(char)0xb0;	//DATA LSB
	evm.midiData[1]=0x26;
	evm.midiData[2]=0;		//cents

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);

	evm.midiData[0]=(char)0xb0;	//RPN Reset
	evm.midiData[1]=0x64;
	evm.midiData[2]=0x7f;

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);

	evm.midiData[0]=(char)0xb0;	//RPN Reset
	evm.midiData[1]=0x65;
	evm.midiData[2]=0x7f;

	evs.events[0]=(VstEvent*)&evm;
	sendVstEventsToHost(&evs);
}



void ChipArp::processReplacing(float **inputs,float **outputs,VstInt32 sampleFrames)
{
	VstEvents evs;
	VstMidiEvent evm;
	char *midiData;
	VstInt32 i,status,note,velocity,bend,arp_dir,low_note,high_note;
	float delta;
	bool key_release,bend_update;

	bendOutRange=pBendRangeList[(int)(pBendRange*4.99f)];

	SetPitchBendRange(); //better to send the range setup always, to avoid glitches when a synth VSTi gets replaced

	arp_dir=(int)(pDirection*2.99f);

	evs.numEvents=1;

	evm.byteSize=sizeof(evm);
	evm.deltaFrames=0;
	evm.type=kVstMidiType;

	arpCounterAdd=(pBPM*MAX_BPM)/60.0f*(float)pSpeedSteps[(int)(pSpeed*7.99f)]/sampleRate;

	bend_update=false;

	while(--sampleFrames>=0)
	{
		key_release=false;

		for(i=0;i<(int)MidiInQueue.size();++i)
		{
			if(MidiInQueue[i].deltaFrames>=0)
			{
				--MidiInQueue[i].deltaFrames;

				if(MidiInQueue[i].deltaFrames<=0)	//new message
				{
					MidiInQueue[i].deltaFrames=0;

					midiData=MidiInQueue[i].midiData;

					status=midiData[0]&0xf0;

					switch(status)
					{
					case 0x80:	//note off
					case 0x90:	//note on
						{
							note=midiData[1]&0x7f; 

							velocity=midiData[2]&0x7f; 

							if(status==0x80) velocity=0;	//note off

							if(velocity&&!MidiKeyState[note])
							{
								MidiKeyState[note]=velocity;

								++MidiKeyDownCount;
							}

							if(!velocity&&MidiKeyState[note])
							{
								MidiKeyState[note]=0;

								--MidiKeyDownCount;

								if(!MidiKeyDownCount)
								{
									key_release=true;
								}
							}
						}
						break;

					case 0xe0:	//catch incoming pitch bend messages
						{
							bendIn=(float)((((midiData[1]&0x7f)|((midiData[2]&0x7f)<<7))-0x2000)*bendInRange/8192.0f);

							bend_update=true;
						}
						break;

					default:	//relay all other messages
						{
							evs.events[0]=(VstEvent*)&MidiInQueue[i];

							sendVstEventsToHost(&evs);

							if(status==0xb0)	//control change
							{
								if(midiData[1]==0x64) MidiRPNLSB =midiData[2]&0x7f;
								if(midiData[1]==0x65) MidiRPNMSB =midiData[2]&0x7f;
								if(midiData[1]==0x26) MidiDataLSB=midiData[2]&0x7f;

								if(midiData[1]==0x06)
								{
									MidiDataMSB=midiData[2]&0x7f;

									if(MidiRPNLSB==0&&MidiRPNMSB==0) bendInRange=(float)MidiDataMSB*.5f;
								}

								if(midiData[1]>=0x7b)	//all notes off and mono/poly mode get redirected, but also releases all notes
								{
									memset(MidiKeyState,0,sizeof(MidiKeyState));

									MidiKeyDownCount=0;
								}
							}
						}
					}
				}
			}
		}

		low_note=-1;

		for(i=0;i<128;++i)
		{
			if(MidiKeyState[i])
			{
				low_note=i;
				break;
			}
		}

		high_note=-1;

		for(i=127;i>=0;--i)
		{
			if(MidiKeyState[i])
			{
				high_note=i;
				break;
			}
		}

		if(key_release)	//all keys were released in current tick
		{
			if(arpBaseNote>=0)	//release latest base note, if any
			{
				evm.midiData[0]=(char)0x80;	//note off
				evm.midiData[1]=arpBaseNote;
				evm.midiData[2]=0x00;

				evs.events[0]=(VstEvent*)&evm;
				sendVstEventsToHost(&evs);

				arpBaseNote=-1;
				arpCurrentNote=-1;
			}
		}
		
		if(low_note>=0||high_note>=0) //a key has been pressed in current tick
		{
			if(arpBaseNote<0)	//if no notes playing, start playing the lowest note and reset arpeggio
			{
				if(arp_dir==1) arpCurrentDir=-1; else arpCurrentDir=1;

				arpBaseNote=arpCurrentDir>0?low_note:high_note;
				arpCurrentNote=arpBaseNote;
				arpCounterAcc=0;

				evm.midiData[0]=(char)0xe0;	//reset pitch bend
				evm.midiData[1]=8192&0x7f;
				evm.midiData[2]=(8192>>7)&0x7f;

				evs.events[0]=(VstEvent*)&evm;
				sendVstEventsToHost(&evs);

				evm.midiData[0]=(char)0x90;	//note on
				evm.midiData[1]=arpBaseNote;
				evm.midiData[2]=MidiKeyState[arpBaseNote];	//take velocity of the lowest note

				evs.events[0]=(VstEvent*)&evm;
				sendVstEventsToHost(&evs);
			}
		}

		delta=bendIn;

		if(arpCurrentNote>=0)
		{
			arpCounterAcc+=arpCounterAdd;

			if(arpCounterAcc>=1.0f)
			{
				arpCounterAcc-=1.0f;

				//find next active note

				for(i=0;i<128;++i)
				{
					arpCurrentNote=(arpCurrentNote+arpCurrentDir)&127;

					if(MidiKeyState[arpCurrentNote]>0) break;
				}

				//reverse direction for the up/down mode

				if(arp_dir==2)
				{
					if(arpCurrentNote==low_note)  arpCurrentDir= abs(arpCurrentDir);
					if(arpCurrentNote==high_note) arpCurrentDir=-abs(arpCurrentDir);
				}

				//try to pitch bend from base to current note

				bend_update=true;
			}

			delta+=(arpCurrentNote-arpBaseNote);
		}

		if(bend_update)
		{
			bend_update=false;

			if(delta<-bendOutRange) delta=-bendOutRange;
			if(delta> bendOutRange) delta= bendOutRange;

			bend=(VstInt32)(8192.0f+(8192.0f/bendOutRange*delta));

			evm.midiData[0]=(char)0xe0;	//pitch bend
			evm.midiData[1]= bend    &0x7f;
			evm.midiData[2]=(bend>>7)&0x7f;

			evs.events[0]=(VstEvent*)&evm;

			sendVstEventsToHost(&evs);
		}
	}

	MidiInQueue.clear();
}
