#include <cmath>
#include <vector>

using namespace std;

#include "audioeffectx.h"


#define PLUGIN_NAME				"ChipArp MIDI"
#define PLUGIN_VENDOR			"v0.2 19.04.17 Shiru"
#define PLUGIN_UID				'carp'

#define NUM_INPUTS				0
#define NUM_OUTPUTS				0
#define NUM_PROGRAMS			1

#define	MAX_BPM					240



enum {
	pIdBPM=0,
	pIdSpeed,
	pIdDirection,
	pIdBendRange,
	NUM_PARAMS
};



const VstInt32 pSpeedSteps[8]={
	1,2,4,8,16,32,64,128
};

const char* const pArpDirNames[3]={
	"Up",
	"Down",
	"Up/Down"
};

const float pBendRangeList[5]={
	60,48,36,24,12
};

const char *pBendRangeNames[5]={
	"60 (Expanded)",
	"48 (Expanded)",
	"36 (Expanded)",
	"24 (Expanded)",
	"12 (Standard)"
};



class ChipArp:public AudioEffectX
{
public:

	ChipArp(audioMasterCallback audioMaster);

	~ChipArp();

	virtual void setParameter(VstInt32 index,float value);
	virtual float getParameter(VstInt32 index);
	virtual void getParameterLabel(VstInt32 index,char *label);
	virtual void getParameterDisplay(VstInt32 index,char *text);
	virtual void getParameterName(VstInt32 index,char *text); 

	virtual VstInt32 getProgram();
	virtual void setProgram(VstInt32 program);
	virtual void getProgramName(char* name);
	virtual void setProgramName(char* name);

	virtual bool getEffectName(char *name) { strcpy(name,PLUGIN_NAME); return true; }
	virtual bool getVendorString(char *text) { strcpy(text,PLUGIN_VENDOR); return true; }
	virtual VstInt32 getVendorVersion() { return 1000; } 

	VstInt32 canDo(char* text);
	VstInt32 getNumMidiInputChannels(void);
	VstInt32 getNumMidiOutputChannels(void);

	VstInt32 processEvents(VstEvents* ev);
	void processReplacing(float**inputs,float **outputs,VstInt32 sampleFrames);

protected:

	void SetPitchBendRange(void);

	float pBPM;
	float pSpeed;
	float pDirection;
	float pBendRange;

	char programName[32];

	unsigned char MidiKeyState[128];

	VstInt32 MidiKeyDownCount;

	vector<VstMidiEvent> MidiInQueue;

	VstInt32 MidiRPNLSB;
	VstInt32 MidiRPNMSB;
	VstInt32 MidiDataLSB;
	VstInt32 MidiDataMSB;

	VstInt32 arpBaseNote;
	VstInt32 arpCurrentNote;
	VstInt32 arpCurrentDir;

	float bendOutRange;

	float bendIn;
	float bendInRange;

	float arpCounterAdd;
	float arpCounterAcc;
};
